
#include "is.h"
#include <cutil.h>

/*****************************************************************/
/*************             R  A  N  K 1           ****************/
/*****************************************************************/

/* Initialization kernel */
__global__ void rank1_kernel_init( 
	int mpi_rank,
	int iteration,
	INT_TYPE *key_array,
	INT_TYPE *bucket_size,
	INT_TYPE *bucket_size_totals,
	INT_TYPE *process_bucket_distrib_ptr1,
	INT_TYPE *process_bucket_distrib_ptr2,
	INT_TYPE2 *test_index_array
	)
{

    INT_TYPE    i;
	
/* 	============================ 
	Determine CUDA environment 
	============================ */ 
	
	INT_TYPE	threadId = threadIdx.x;
	INT_TYPE	blockId = blockIdx.x;
	INT_TYPE	blockSize = blockDim.x;
	INT_TYPE	blockCount = gridDim.x;
	INT_TYPE	cudaGlobalId = ((blockId*blockSize) + threadId);
	INT_TYPE	cudaGlobalSize = (blockSize*blockCount);
	INT_TYPE	BUCKETS_PER_CUDA_THREAD = NUM_BUCKETS / cudaGlobalSize;
	if(BUCKETS_PER_CUDA_THREAD < 1)
		BUCKETS_PER_CUDA_THREAD = 1;

/*  	============================
	Iteration alteration of keys by single thread #0 
	============================ */  
	
    if( (mpi_rank == 0 ) && (cudaGlobalId == 0))
    {
      key_array[iteration] = iteration;
      key_array[iteration+MAX_ITERATIONS] = MAX_KEY - iteration;
    }
	
	__syncthreads();
	__threadfence();

/*  	============================
	Initialize working arrays
	============================ */	
	
	if(cudaGlobalId < NUM_BUCKETS){
	
		for(i = (cudaGlobalId * BUCKETS_PER_CUDA_THREAD); i < ((cudaGlobalId+1) * BUCKETS_PER_CUDA_THREAD); i++){
			bucket_size[i] = 0;
			bucket_size_totals[i] = 0;
			process_bucket_distrib_ptr1[i] = 0;
			process_bucket_distrib_ptr2[i] = 0;
		}
		
	}
	
	__syncthreads();
	__threadfence();
	
	if(cudaGlobalId < TEST_ARRAY_SIZE){
	    bucket_size[NUM_BUCKETS + cudaGlobalId] = 0;
        bucket_size_totals[NUM_BUCKETS + cudaGlobalId] = 0;
		process_bucket_distrib_ptr1[NUM_BUCKETS + cudaGlobalId] = 0;
		process_bucket_distrib_ptr2[NUM_BUCKETS + cudaGlobalId] = 0;
		
		if( (test_index_array[cudaGlobalId]/NUM_KEYS) == mpi_rank )
            bucket_size[NUM_BUCKETS+cudaGlobalId] = key_array[test_index_array[cudaGlobalId] % NUM_KEYS];
	}
	
	__syncthreads();
	__threadfence();
}

/* Key counting kernel */
__global__ void rank1_kernel_countkeys( 
	INT_TYPE *key_array,
	INT_TYPE *bucket_size
	)
{

    INT_TYPE    i;
	INT_TYPE    shift = MAX_KEY_LOG_2 - NUM_BUCKETS_LOG_2;

/* 	============================ 
	Determine CUDA environment 
	============================ */ 
	
	INT_TYPE	threadId = threadIdx.x;
	INT_TYPE	blockId = blockIdx.x;
	INT_TYPE	blockSize = blockDim.x;
	INT_TYPE	blockCount = gridDim.x;
	INT_TYPE	cudaGlobalId = ((blockId*blockSize) + threadId);
	INT_TYPE	cudaGlobalSize = (blockSize*blockCount);
	INT_TYPE	KEYS_PER_CUDA_THREAD = NUM_KEYS / cudaGlobalSize;

		
/*  	============================
	Determine the number of keys in each bucket 
	============================ */	

	for( i=(cudaGlobalId * KEYS_PER_CUDA_THREAD); i<((cudaGlobalId + 1) * KEYS_PER_CUDA_THREAD); i++ ){     
		// Save shifted key from global memory
		int key_shifted = key_array[i] >> shift;
		
		// Increment this bucket_size position
		atomicAdd(&bucket_size[key_shifted], 1);
	}
	
	__syncthreads();
	__threadfence();

}

/* Bucket size kernel */
__global__ void rank1_kernel_bucketsizes( 
	INT_TYPE *bucket_size,
	INT_TYPE *bucket_ptrs
	)
{

    INT_TYPE    i;
	
/* 	============================ 
	Determine CUDA environment 
	============================ */ 
	
	INT_TYPE	threadId = threadIdx.x;
	INT_TYPE	blockId = blockIdx.x;
	INT_TYPE	blockSize = blockDim.x;
	INT_TYPE	blockCount = gridDim.x;
	INT_TYPE	cudaGlobalId = ((blockId*blockSize) + threadId);
	INT_TYPE	cudaGlobalSize = (blockSize*blockCount);
	INT_TYPE	BUCKETS_PER_CUDA_THREAD = NUM_BUCKETS / cudaGlobalSize;
	if(BUCKETS_PER_CUDA_THREAD < 1)
		BUCKETS_PER_CUDA_THREAD = 1;

/*  	============================
	Accumulative bucket sizes are the bucket pointers 
	============================ */

	// Delcare shared array - can't use more than 1 block for this!
	// extern __shared__ INT_TYPE bucket_ptrs_shared[];
	__shared__ INT_TYPE bucket_ptrs_shared[NUM_BUCKETS];
	
	// Node 0 calculates
	if(cudaGlobalId == 0){
	
		bucket_ptrs_shared[0] = 0;
		for( i=1; i< NUM_BUCKETS; i++ )  
			bucket_ptrs_shared[i] = bucket_ptrs_shared[i-1] + bucket_size[i-1];
	
	}
	
	__syncthreads();
	
/*  	============================
	Copy back into global mem
	============================ */

	if(cudaGlobalId < NUM_BUCKETS){
	
		for(i = (cudaGlobalId * BUCKETS_PER_CUDA_THREAD); i < ((cudaGlobalId+1) * BUCKETS_PER_CUDA_THREAD); i++){
			bucket_ptrs[i] = bucket_ptrs_shared[i];
		}
		
	}
	
	__syncthreads();
	
}

/* Sorting kernel */
__global__ void rank1_kernel_sort( 
	INT_TYPE *key_array,
	INT_TYPE *bucket_size,
	INT_TYPE *bucket_ptrs,
	INT_TYPE *key_buff1
	)
{

    INT_TYPE    i;
	INT_TYPE    shift = MAX_KEY_LOG_2 - NUM_BUCKETS_LOG_2;
	
/* 	============================ 
	Determine CUDA environment 
	============================ */ 
	
	INT_TYPE	threadId = threadIdx.x;
	INT_TYPE	blockId = blockIdx.x;
	INT_TYPE	blockSize = blockDim.x;
	INT_TYPE	blockCount = gridDim.x;
	INT_TYPE	cudaGlobalId = ((blockId*blockSize) + threadId);
	INT_TYPE	cudaGlobalSize = (blockSize*blockCount);
	INT_TYPE	KEYS_PER_CUDA_THREAD = NUM_KEYS / cudaGlobalSize;


/*  	============================
	Sort into appropriate bucket 
	============================ */

	for( i=(cudaGlobalId * KEYS_PER_CUDA_THREAD); i<((cudaGlobalId + 1) * KEYS_PER_CUDA_THREAD); i++ ){   
		
		// Save key from global memory and determine shifted value
		int key = key_array[i];
		int key_shifted = key >> shift;
		
		// Increment the bucket_ptrs value at index [key_array[i] >> shift]
		int bucket_ptrs_val = atomicAdd(&bucket_ptrs[key_shifted], 1);

		// Set key_buff1
		atomicExch(&key_buff1[bucket_ptrs_val], key);
		
	}

}



/*****************************************************************/
/*************             R  A  N  K 2           ****************/
/*****************************************************************/

__global__ void rank2_kernel( 
	int mpi_node_count,
	INT_TYPE *bucket_size,
	INT_TYPE *bucket_size_totals,
	INT_TYPE *process_bucket_distrib_ptr1,
	INT_TYPE *process_bucket_distrib_ptr2,
	INT_TYPE *send_displ,
	INT_TYPE *send_count	
	)
{

    INT_TYPE    i;
	INT_TYPE2   bucket_sum_accumulator, j;
    INT_TYPE    local_bucket_sum_accumulator;

/* 	============================ 
	Determine CUDA environment 
	============================ */ 
	
	INT_TYPE	threadId = threadIdx.x;
	INT_TYPE	blockId = blockIdx.x;
	INT_TYPE	blockSize = blockDim.x;
	INT_TYPE	blockCount = gridDim.x;
	INT_TYPE	cudaGlobalId = ((blockId*blockSize) + threadId);
	INT_TYPE	cudaGlobalSize = (blockSize*blockCount);
	INT_TYPE	BUCKETS_PER_CUDA_THREAD = NUM_BUCKETS / cudaGlobalSize;
	if(BUCKETS_PER_CUDA_THREAD < 1)
		BUCKETS_PER_CUDA_THREAD = 1;


/*  Determine Redistibution of keys: accumulate the bucket size totals 
    till this number surpasses NUM_KEYS (which the average number of keys
    per processor).  Then all keys in these buckets go to processor 0.
    Continue accumulating again until supassing 2*NUM_KEYS. All keys
    in these buckets go to processor 1, etc.  This algorithm guarantees
    that all processors have work ranking; no processors are left idle.
    The optimum number of buckets, however, does not result in as high
    a degree of load balancing (as even a distribution of keys as is
    possible) as is obtained from increasing the number of buckets, but
    more buckets results in more computation per processor so that the
    optimum number of buckets turns out to be 1024 for machines tested.
    Note that process_bucket_distrib_ptr1 and ..._ptr2 hold the bucket
    number of first and last bucket which each processor will have after   
    the redistribution is done.                                          */

/* 	============================ 
	Shared memory version
	============================ */ 

	__shared__ int send_displ_shared[NUM_PROCS];
	__shared__ int send_count_shared[NUM_PROCS];
	__shared__ INT_TYPE process_bucket_distrib_ptr1_shared[NUM_BUCKETS];
	__shared__ INT_TYPE process_bucket_distrib_ptr2_shared[NUM_BUCKETS];

	// First thread performs calcs
	if(cudaGlobalId == 0){
	
		bucket_sum_accumulator = 0;
		local_bucket_sum_accumulator = 0;
	
		send_displ_shared[0] = 0;
		process_bucket_distrib_ptr1_shared[0] = 0;
		for( i=0, j=0; i<NUM_BUCKETS; i++ )  
		{
			bucket_sum_accumulator       += bucket_size_totals[i];
			local_bucket_sum_accumulator += bucket_size[i];
			if( bucket_sum_accumulator >= (j+1)*NUM_KEYS )  
			{
				send_count_shared[j] = local_bucket_sum_accumulator;
				if( j != 0 )
				{
					send_displ_shared[j] = send_displ_shared[j-1] + send_count_shared[j-1];
					process_bucket_distrib_ptr1_shared[j] = process_bucket_distrib_ptr2_shared[j-1]+1;
				}
				process_bucket_distrib_ptr2_shared[j] = i;
				j++;
				local_bucket_sum_accumulator = 0;
			}
		}
		
		while( j < mpi_node_count )
		{
			send_count_shared[j] = 0;
			process_bucket_distrib_ptr1_shared[j] = 1;
			j++;
		}
	
	}
	
	__syncthreads();
	
	// All threads copy out process_bucket_distrib_ptr* arrays
	if(cudaGlobalId < NUM_BUCKETS){
	
		for(i = (cudaGlobalId * BUCKETS_PER_CUDA_THREAD); i < ((cudaGlobalId+1) * BUCKETS_PER_CUDA_THREAD); i++){
			process_bucket_distrib_ptr1[i] = process_bucket_distrib_ptr1_shared[i];
			process_bucket_distrib_ptr2[i] = process_bucket_distrib_ptr2_shared[i];
		}
		
	}
	
	// Lower-number threads copy out send_* arrays
	if(cudaGlobalId < NUM_PROCS){
	
		send_count[cudaGlobalId] = send_count_shared[cudaGlobalId];
		send_displ[cudaGlobalId] = send_displ_shared[cudaGlobalId];
	
	}
	
	__syncthreads();
	
}    
   